/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.engstelle;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.NonMutableSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.Konfiguration;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.MessQuerschnitt;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.StrassenSegment;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.VerkehrsstaerkeFs;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungEngstellen;

/**
 * Modelliert eine Baustelle als spezielle Engstelle
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.3 $ / $Date: 2009/09/23 13:55:28 $ / ($Author: Sans $)
 */
public class Baustelle extends Engstelle 
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * Restkapazitt in Fzg/h
   */
  private long _restKapazitaet = -1;
  
  /**
   * Weist ihr Status diese Baustelle als gltig aus?
   */
  private Boolean _statusGueltig = null;
  
  /**
   * Erzeugt eine neue Instanz vom Typ Baustelle
   * 
   * @param verwaltung die Engstellenverwaltung, die fr diese Instanz zustndig ist
   * @param davObjekt das Baustellen-Systemobjekt
   */
  public Baustelle(VerwaltungEngstellen verwaltung, SystemObject davObjekt) 
  {
    super(verwaltung, davObjekt);
        
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.baustellenEigenschaftenErweitert");
    Data data = null;
    
    try
    {
      data = davObjekt.getConfigurationData(atg);
    }
    catch (Exception e)
    {
      return; // Attributgruppe ist (hoffentlich) parametrierend
    }
    if (data != null)
    {  
      Array array = data.getArray("LageGesperrteFahrstreifen");
      for (int i = 0; i < array.getLength(); i += 1)
      {
        _lageGesperrterFahrstreifen.add(array.getItem(i).asTextValue().getText());
      }
      _debug.fine("Fr " + this + " gibt es Daten gesperrter Fahrstreifen");
      
    }
    else
    {
      _debug.fine("Fr " + this + " gibt KEINE Daten gesperrter Fahrstreifen");
    } 
  }
    
  
  /**
   * @return Die Restkapazitt dieser Baustelle oder -1, 
   * wenn die Restkapazitt nicht bestimmt werden konnte
   */
  public long getEngpassKapazitaet()
  {
    return _restKapazitaet;
  }
  
  /**
   * Bestimmt, ob diese Baustelle durch ihren Status als gltig 
   * ausgewiesen wird
   * 
   * @return true, wenn diese Baustelle gltig ist, false wenn sie
   * nicht gltig ist oder ihr Status nicht bestimmte werden kann
   */
  protected boolean isStatusGueltig()
  {
    if (_statusGueltig == null)
    {
      _debug.fine(this + " nicht prfbar da Status nicht spezifiziert");
      return false;
    }
    if (!_statusGueltig.booleanValue())
    {
      _debug.fine("Status von " + this + " weist diese Baustelle als nicht gueltig aus");
      return false;
    }
    return true;
  }
  
  @Override
  public boolean pruefeGueltigkeit(long zeitpunkt)
  {
    return (isStatusGueltig() && super.pruefeGueltigkeit(zeitpunkt));
  }  
  
  /**
   * Bestimmt den Parameter MaxVerkehrsStrkeGesperrteFahrStreifen der 
   * Engstellenverwaltung
   * 
   *  @return den Wert des Parameters
   */
  public int getMaxVerkehrsStaerkeGesperrteFahrStreifen()
  {
    return _verwaltung.getMaxVerkehrsStaerkeGesperrteFahrStreifen();
  }
  
  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnmeldung#anmelden()
   */
  public void anmelden()
  {
    super.anmelden();
    
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.baustellenEigenschaften");
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    DataDescription datenbeschreibung = new DataDescription(atg, asp);
    
    getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung,
        ReceiveOptions.normal(), ReceiverRole.receiver()); 
    
 
    atg = getConnection().getDataModel().getAttributeGroup("atg.baustellenEigenschaftenErweitert");
    
    if (atg.isParameter())
    { 
      datenbeschreibung = new DataDescription(atg, asp);
      getConnection().subscribeReceiver(this, getObjekt(), datenbeschreibung,
          ReceiveOptions.normal(), ReceiverRole.receiver());
    }
  }
  
  /*
   * (Kein Javadoc)
   * @see de.bsvrz.vew.vewallg.empf.EmpfaengerMitAnAbmeldung#abmelden()
   */
  public void abmelden()
  {
    Aspect asp = getConnection().getDataModel().getAspect("asp.parameterSoll");
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.baustellenEigenschaftenErweitert");
    DataDescription datenbeschreibung;
    
    if (atg.isParameter())
    {  
      datenbeschreibung = new DataDescription(atg, asp);
      getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
      bereinigeGesperrteFahrstreifen();
    }
    
    atg = getConnection().getDataModel().getAttributeGroup("atg.baustellenEigenschaften");
    datenbeschreibung = new DataDescription(atg, asp);
    getConnection().unsubscribeReceiver(this, getObjekt(), datenbeschreibung);
    
    super.abmelden();
  }
  
  /*
   * (Kein Javadoc)
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] resultData)
  {
    Lage lageAlt = getLage();
    super.update(resultData);
    boolean updateGesperrteFahrstreifen = (getLage() != lageAlt);    
    
    // Ausgaben
    for (ResultData data : resultData)
    {
      Aspect asp = data.getDataDescription().getAspect();
      AttributeGroup atg = data.getDataDescription().getAttributeGroup();
      short sv = data.getDataDescription().getSimulationVariant();
      
      _debug.finest("Daten empfangen mit " + atg.getPid() + "/" +
          asp.getPid() + "/" + sv  + ":\n" + data);
      
      if (asp.getPid().equals("asp.parameterSoll") && 
          atg.getPid().equals("atg.baustellenEigenschaften"))
      {
        if (data.hasData())
        {
          _debug.finer("Bearbeite Daten mit " + atg.getPid() + "/" +
              asp.getPid() + "/" + sv + ":\n" + data);
          
          Data item = data.getData().getItem("Status");
          String status = item.asTextValue().getText();
          if (status.equals("gltig"))
          {
            _statusGueltig = Boolean.TRUE;
          }
          else
          {
            _statusGueltig = Boolean.FALSE;
          }
          
          item = data.getData().getItem("RestKapazitt");
          _restKapazitaet = item.asUnscaledValue().longValue();
        }
        else
        {
          _statusGueltig = null;
          _restKapazitaet = -1;
          _debug.config("Restkapazitt und Status sind unbestimmt fr Baustelle" + this);
        }
        _debug.fine("Update Status fr " + this + " Wert ist " + 
            ((_statusGueltig == null || !_statusGueltig) ? "\"ungltig\"" :"\"gltig\""));
        _debug.fine("Update Restkapazitt fr " + this + " Wert ist " + _restKapazitaet);
      }
      if (asp.getPid().equals("asp.parameterSoll") && 
          atg.getPid().equals("atg.baustellenEigenschaftenErweitert"))
      {
        _lageGesperrterFahrstreifen.clear();
        if (data.hasData())
        {
          Array array = data.getData().getArray("LageGesperrteFahrstreifen");
          for (int i = 0; i < array.getLength(); i += 1)
          {
            _lageGesperrterFahrstreifen.add(array.getItem(i).asTextValue().getText());
          }
        }
        _debug.fine("Update LageGesperrteFahrstreifen fr " + this + 
            " Anzahl ist " + _lageGesperrterFahrstreifen.size());
        updateGesperrteFahrstreifen = true;
      }
    }
    
    if (updateGesperrteFahrstreifen) 
    {
      bestimmmeGesperrteFahrstreifen();
    }
  }  
  
  /**
   * Instanz von <code>MeldungsBehandlung</code> zur Behandlung von Betriebsmeldungen zur 
   * Fahrstreifenprfung dieser Baustelle
   */
  public final MeldungsBehandlung fahrStreifenMeldungen = new MeldungsBehandlung();
  
  /**
   * Bildet eine Id zur Verwaltung von Betriebsmeldungen bezgl. der Fahrstreifenprfung 
   * fr diese Baustelle
   * 
   * @param fs der referenzierte Fahrstreifen
   * @return die gebildete Id
   */
  public String getIdFahrStreifenMeldung(SystemObject fs)
  {
    return getObjekt().getPid() + "." + fs.getPid();
  }
  
  /**
   * Sammlung der Bezeichner der gesperrten Fahrstreifen fr diese Baustelle ("HFS", "FS1" usw.)
   */
  private Collection<String> _lageGesperrterFahrstreifen = new ArrayList<String>();
  
  /**
   * Read-only Zugriff auf dei Sammlung der Bezeichner der gesperrten Fahrstreifen 
   * fr diese Baustelle
   * 
   * @return Sammlung der Bezeichner der gesperrten Fahrstreifen ("HFS", FS1 usw.)
   */
  public Collection<String> getLageGesperrterFahrstreifen()
  {
    return Collections.unmodifiableCollection(_lageGesperrterFahrstreifen);
  }
  
  /**
   * Enthlt fr jeden gesperrten Fahrstreifen eine Instanz vom Typ <code>VerkehrsstaerkeFs</code>,
   * Schlssel ist das Fahrstreifen-Systemobjekt.
   */
  private Map<SystemObject, VerkehrsstaerkeFs> _gesperrteFahrstreifen =
    new HashMap<SystemObject, VerkehrsstaerkeFs>();
  
  /**
   * Leert die Map <code>_gesperrteFahrstreifen</code> und gibt mit ihren Eintrgen
   * evtl. verbundene Ressourcen frei.
   */
  protected void  bereinigeGesperrteFahrstreifen()
  {
    synchronized(_gesperrteFahrstreifen) 
    {
      for (SystemObject fs:_gesperrteFahrstreifen.keySet())
      {
        VerkehrsstaerkeFs vs = _gesperrteFahrstreifen.get(fs);
        vs.dispose();
        fahrStreifenMeldungen.setMeldungInaktiv(getIdFahrStreifenMeldung(fs));
      }
      _gesperrteFahrstreifen.clear();
    }
  }
  
  /**
   * Initialisiert die Map <code>_gesperrteFahrstreifen</code>
   */
  protected void  bestimmmeGesperrteFahrstreifen()
  {
    Lage lage = getLage();
    
    synchronized(_gesperrteFahrstreifen)
    {
      bereinigeGesperrteFahrstreifen();
      
      if (lage != null)
      {
        for (int i = 0; i < lage.strassenSegmente.getLength(); i++)
        {
          SystemObject ss = 
            lage.strassenSegmente.getItem(i).asReferenceValue().getSystemObject();
          List<MessQuerschnitt> messQuerschnitte =
            Konfiguration.getInstanz().getMessQuerschnitteVonStrassenSegment(ss.getPid());
          
          if (messQuerschnitte != null)
          {
            // BUGFIX bugzilla Bug 46: Alle MQ, die nicht vom Typ typ.messQuerschnitt sind,
            // werden ausgeschlossen
            ListIterator<MessQuerschnitt> iter = messQuerschnitte.listIterator();
            while (iter.hasNext())
            {
              MessQuerschnitt m = iter.next();
              if (!m.getSystemObjekt().isOfType("typ.messQuerschnitt"))
              {
                iter.remove();
              }
            }
            
            if (i == 0)
            {
              // Erstes Strassensegment -> entferne MQ vor OffsetVor des Strassensegments
              long offset = (long)lage.offsetVor;
              ListIterator<MessQuerschnitt> it = messQuerschnitte.listIterator();
              while (it.hasNext())
              {
                MessQuerschnitt m = it.next();
                if (m.getOffset() < offset)
                {
                  it.remove();
                }
              }
            }
            if (i == lage.strassenSegmente.getLength() - 1)
            {
              // Letztes Strassensegment -> entferne MQ nach OffsetNach des Strassensegments
              StrassenSegment segment = 
                Konfiguration.getInstanz().getStrassenSegment(ss.getPid());
              long offset = segment.getLaenge() - (long)lage.offsetNach;
              ListIterator<MessQuerschnitt> it = messQuerschnitte.listIterator();
              while (it.hasNext())
              {
                MessQuerschnitt m = it.next();
                if (m.getOffset() > offset)
                {
                  it.remove();
                }
              }
            }
          
            for (MessQuerschnitt mq: messQuerschnitte)
            {
              ConfigurationObject co = (ConfigurationObject)mq.getSystemObjekt();
              NonMutableSet set = co.getNonMutableSet("FahrStreifen");
              if (set != null)
              {   
                List<SystemObject> fahrStreifen = set.getElements();
                
                for (SystemObject fs: fahrStreifen)
                {
                  String fsLage = getFahrstreifenLage(fs);
                  
                  for (String gfsLage: getLageGesperrterFahrstreifen())
                  {
                    if (gfsLage.equals(fsLage))
                    {
                      _gesperrteFahrstreifen.put(fs, new VerkehrsstaerkeFs(getConnection(), fs.getPid()));
                    }
                  }
                }
              }  
            }
          }
        }
      }
    }
    _debug.config("Baustelle " + this + " hat " +_gesperrteFahrstreifen.size() +
        " Mestellen auf gesperrten Fahrstreifen");
  }
   
  /**
   * Bestimmt die Lage eines Fahrstreifens
   * 
   * @param fs der Fahrstreifen, dessen Lage bestimmt werden soll
   * @return die Lage des Fahrstreifens
   * @throws IllegalArgumentException falls fs nicht vom Typ "typ.fahrStreifen" ist
   */
  private String getFahrstreifenLage(SystemObject fs)
  {
    if (!fs.isOfType("typ.fahrStreifen"))
    {
      throw new IllegalArgumentException("Objekt vom Typ typ.fahrStreifen erwartet");
    }
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.fahrStreifen");
    String lage = fs.getConfigurationData(atg).getItem("Lage").asTextValue().getText();
    return lage == null ? "" : lage;
  }
  
  /**
   * Fhrt die Fahrstreifenprfung fr diese Baustelle durch.
   * Die Verkehrstrke auf allen gesperrten Fahrstreifen wird mit dem Parameter
   * <code>MaxVerkehrsStaerkeGesperrteFahrStreifen</code> verglichen. Alle Fahrstreifen,
   * auf denen die Verkehrsstrke den Wert des Parameters berschreitet, werden 
   * zurckgemeldet.
   * 
   * @return Eine Sammlung von <code>ErgebnisFsPruefung</code> der Fahrstreifen-Objekte (Mestellen auf 
   * gesperrten Fahrstriefen), auf denen die Verkehrsstrke den Wert des Parameters 
   * <code>MaxVerkehrsStaerkeGesperrteFahrStreifen</code> berschreitet. 
   * Eine leere Sammlung bedeutet dabei ein positives, eine nichtleere Sammlung ein 
   * negatives Ergebnis der Plausibilittsprfung. 
   * @throws IllegalStateException wenn die Prfung nicht durchgefhrt werden kann, aufgrund
   * etwa nichtversorgter Parameter, Baustelle nicht gltig etc.
   */
  public Collection<ErgebnisFsPruefung> pruefeGesperrteFahrstreifen()
    throws IllegalStateException
    {
    int maxVerkehrsStaerkeGesperrteFahrStreifen = getMaxVerkehrsStaerkeGesperrteFahrStreifen();
      if (maxVerkehrsStaerkeGesperrteFahrStreifen < 0)
      {
        throw new IllegalStateException(this + " ist nicht prfbar " +  
          "da Parameter \"MaxVerkehrsStaerkeGesperrteFahrStreifen\" nicht versorgt");
      }
          
      long currentTime = getConnection().getTime();

      if (!pruefeGueltigkeit(currentTime))
      {
        throw new IllegalStateException(this + " ist nicht prfbar " +  
          "da (momentan) nicht gltig");
      }
   
      if (!pruefeLage())
      {
        throw new IllegalStateException(this + " ist nicht prfbar " +  
          "da Lage unspezifiziert");
      }
      
      Collection<ErgebnisFsPruefung> result = null;
      
      Collection<SystemObject> fahrStreifen = _gesperrteFahrstreifen.keySet();
      _debug.config("Prfe " + fahrStreifen.size() + " Mestellen gesperrter Fahrstreifen der Engstelle " + this);
      Iterator<SystemObject> it = fahrStreifen.iterator();
            
      while(it.hasNext())
      {
        try
        {
          SystemObject fs = it.next();
          VerkehrsstaerkeFs verkehrsStaerke = _gesperrteFahrstreifen.get(fs);

          _debug.fine("Auf Fahrstreifen-Mestelle " + fs.getName() + " ("+ getFahrstreifenLage(fs) + 
              ") von " + this + " ist die Verkehrsstrke " + (verkehrsStaerke.getVerkehrsstaerkeQKfz() / 60) + 
              "Fzg/min - Plausibel = " + verkehrsStaerke.isQKfzPlausibel());
          
          if (verkehrsStaerke.isQKfzPlausibel())
          {
            long vs = verkehrsStaerke.getVerkehrsstaerkeQKfz() / 60; 
            if (vs > maxVerkehrsStaerkeGesperrteFahrStreifen)
            {
              if (result == null)
              {
                result = new ArrayList<ErgebnisFsPruefung>();
              }
              result.add(new ErgebnisFsPruefung(fs, vs));
              _debug.info("Auf Fahrstreifen-Mestelle " + fs.getName() + " ("+ getFahrstreifenLage(fs) + 
                  ", gesperrt) von " + this + " ist die Verkehrsstrke mit " + vs +
                  "Fzg/min grsser als die zulssigen " + maxVerkehrsStaerkeGesperrteFahrStreifen + "Fzg/min");
            }
            else
            {
              fahrStreifenMeldungen.setMeldungInaktiv(getIdFahrStreifenMeldung(fs));
            }
          }
          else
          {
            _debug.fine("An Fahrstreifen-Mestelle " + fs.getName() + " ist der Messwert nicht plausibel");
            continue;
          }
        }
        catch (ConcurrentModificationException e)
        {
          _debug.config("Prfe " + fahrStreifen.size() + "Mestellen gesperrter Fahrstreifen der Engstelle " + this);
          it = fahrStreifen.iterator();
        }
      }
      
      if (result == null)
      {  
        result = Collections.emptyList();
      }  
      
      return result;
  }
}
