/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.engstelle;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.InneresStrassenSegment;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.Konfiguration;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.MessQuerschnitt;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.StrassenKnoten;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.StrassenSegment;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.Verkehrsstaerke;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungEngstellen;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungStaus;

/**
 * Modelliert eine Engstelle als spezielle Situation
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.2 $ / $Date: 2008/11/24 16:05:40 $ / ($Author: Sans $)
 */
public abstract class Engstelle extends Situation
  implements VerwaltungStaus.BereinigungsListener
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();
  
  /**
   * Die Engstellenverwaltung, die fr diese Instanz zustndig ist
   */
  protected final VerwaltungEngstellen _verwaltung;
  
  /**
   * Konstruktor fr den abstrakten Typ Engstelle
   * 
   * @param verwaltung die Engstellenverwaltung, die fr diese Instanz zustndig ist
   * @param davObjekt das Engstellen-Systemobjekt
   */
  protected Engstelle(VerwaltungEngstellen verwaltung,
      SystemObject davObjekt)
  {
    super(verwaltung.getConnection(), davObjekt);
    _verwaltung = verwaltung;
  }
  
  /**
   * Bestimmt die (Engpass-) Kapazitt dieser Engstelle 
   * 
   * @return Die Kapazitt in Fzg/h oder einen Wert <0, falls die
   * Kapazizt nicht bestimmt werden konnte.
   */
  public abstract long getEngpassKapazitaet();
  
  /**
   * Bestimmt den Parameter FaktorMaxVerkehrsStrkeStromAbwrts der 
   * Engstellenverwaltung
   * 
   *  @return den Wert des Parameters
   */
  public float getFaktorVerkehrsStaerkeStromAbwaerts()
  {
    return _verwaltung.getFaktorVerkehrsStaerkeStromAbwaerts();
  }
    
  /**
   * Bestimmt den Parameter MaxLngeStauberBaustellenEnde der 
   * Engstellenverwaltung
   * 
   *  @return den Wert des Parameters
   */
  public long getMaxLaengeStauUeberBaustellenEnde()
  {
    return _verwaltung.getMaxLaengeStauUeberBaustellenEnde();
  }
  
  /**
   * Fhrt die Stauprfung fr diese Engstelle durch.
   * Wenn ein Stau auf dem letzten Strassensegment der Engstelle stattfindet, wird dessen 
   * Lnge ber das Ender der Engstelle hinaus mit dem Parameter 
   * <code>MaxLaengeStauUeberBaustellenEnde</code> verglichen. Alle Staus,
   * die mehr als der Wert des Parameters angibt ber das Engstellenende hinausragen, werden 
   * zurckgemeldet.
   * 
   * @return Eine Sammlung von <code>ErgebnisStauPruefung</code> der Staus, die lnger als der Wert 
   * des Parameters <code>MaxLaengeStauUeberBaustellenEnde</code> angibt ber das Engstellenende hinausragen. 
   * Eine leere Sammlung bedeutet dabei ein positives, eine nichtleere Sammlung ein 
   * negatives Ergebnis der Plausibilittsprfung. 
   * @throws IllegalStateException wenn die Prfung nicht durchgefhrt werden kann, aufgrund
   * etwa nichtversorgter Parameter, Baustelle nicht gltig etc.
   */
  public synchronized Collection<ErgebnisStauPruefung> pruefeStaus()
    throws IllegalStateException
  {
    long maxLaengeStauUeberBaustellenEnde = getMaxLaengeStauUeberBaustellenEnde();
    if (maxLaengeStauUeberBaustellenEnde < 0)
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
      "da Parameter \"MaxLaengeStauUeberBaustellenEnde\" nicht versorgt");
    }
        
    long currentTime = getConnection().getTime();
   
    if (!pruefeGueltigkeit(currentTime))
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
        "da (momentan) nicht gltig");
    }
    
    if (!pruefeLage())
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
      "da Lage unspezifiziert");
    }
    
    Collection<ErgebnisStauPruefung> result = null;
    SystemObject letztesSegment = getLetztesStrassenSegment();
    // float laengeLetztesSegment = getLaengeStrassenSegment(letztesSegment);
    // float endeLetztesSegment = laengeLetztesSegment - getLage().offsetNach;
    float endeLetztesSegment = getLage().offsetNach;
    float anfangLetztesSegment = (getLage().strassenSegmente.getLength() > 1) ? 
        0 : getLage().offsetVor;
    
    _debug.config("Prfe Staus der Engstelle " + this + 
        " - das letzte Strassensegment " + letztesSegment.getName() + 
        " wird von " + anfangLetztesSegment + "m bis " + endeLetztesSegment + 
        "m benutzt");
    
    Iterator<Situation> it = _verwaltung.getVerwaltungsStaus().getVerwalteteObjekte().iterator();
    while (it.hasNext())
    {
      try
      {
        Situation stau = it.next();
        synchronized(stau)
        {
          if (!stau.pruefeGueltigkeit(currentTime))
          {
            continue;
          }
           
          if (!stau.pruefeLage())
          {
            continue;
          }
          
          Array segmente = stau.getLage().strassenSegmente;
          boolean stauGefunden = false;
          float stauLaengeNachEngstelle = 0;
          
          _debug.finer("Prfe Stau " + stau + 
              " mit " + segmente.getLength() + " Strassensegment(en)");
          
          for (int i = 0; i < segmente.getLength(); i+=1)
          {
            Data item = segmente.getItem(i);
            SystemObject segment = item.asReferenceValue().getSystemObject();
            _debug.finest("Prfe Stau-Strassensegment " + segment.getName()); 
          
            if (letztesSegment.getId() == segment.getId()) 
            {   
              // Falls Endsegment auch im Stau vorkommt
              // Setze die Variable stauLaengeNachEngstelle auf den Anfangspunkt des Staus
              // Korrigiere ggf. Offset
              float stauAnfangSegment = (i == 0) ? stau.getLage().offsetVor : 0;
//              float stauEndeSegment = laengeLetztesSegment - ((i == segmente.getLength() - 1) ? 
//                  stau.getLage().offsetNach : 0);
              
              float stauEndeSegment = ((i == segmente.getLength() - 1) ? stau.getLage().offsetNach : getLaengeStrassenSegment(segment));
              
              if (stauAnfangSegment < endeLetztesSegment &&
                  stauEndeSegment >= endeLetztesSegment)
              {
                stauGefunden = true;
                
//                float laenge = endeLetztesSegment - ((anfangLetztesSegment > stauAnfangSegment) ?
//                    anfangLetztesSegment : stauAnfangSegment);
                stauLaengeNachEngstelle = stauEndeSegment - endeLetztesSegment; 

                _debug.config("Stau " + stau + " berlappt Engstelle " + this +
                    " am Ende ihres Endsegments " + letztesSegment.getName() + 
                    " um " + stauLaengeNachEngstelle + "m");
              }
            }
            if (stauGefunden)
            {
              if (letztesSegment.getId() != segment.getId())
              {  
                stauLaengeNachEngstelle += getLaengeStrassenSegment(segment);
                if (i == segmente.getLength() - 1)
                {
                  stauLaengeNachEngstelle -= stau.getLage().offsetNach;
                }
              }  
              if (stauLaengeNachEngstelle > maxLaengeStauUeberBaustellenEnde)
              {
                break;
              }
            }
                      
          }
          if (stauGefunden && stauLaengeNachEngstelle > maxLaengeStauUeberBaustellenEnde)
          {
            if (result == null)
            {
              result = new ArrayList<ErgebnisStauPruefung>();
            }
            result.add(new ErgebnisStauPruefung(stau.getObjekt(), stauLaengeNachEngstelle));
            _debug.info("Stau " + stau + " geht mehr als " + 
                maxLaengeStauUeberBaustellenEnde + "m (" + 
                stauLaengeNachEngstelle + "m) ber das Ende von " +
                this + " hinaus");
          }
          else
          {
            setStauMeldungInaktiv(stau.getObjekt());
          }
        }
      }
      catch(ConcurrentModificationException e)
      {
        it = _verwaltung.getVerwaltungsStaus().getVerwalteteObjekte().iterator();
        _debug.fine("nderung der Staumenge whrend Prfung");
      }
    }  
    
    if (result == null)
    {  
      result = Collections.emptyList();
    }  
    
    return result;
  }

  public synchronized Collection<SystemObject> pruefeStausAlt()
  throws IllegalStateException
{
  long maxLaengeStauUeberBaustellenEnde = getMaxLaengeStauUeberBaustellenEnde();
  if (maxLaengeStauUeberBaustellenEnde < 0)
  {
    throw new IllegalStateException(this + " ist nicht prfbar " +  
    "da Parameter \"MaxLaengeStauUeberBaustellenEnde\" nicht versorgt");
  }
      
  long currentTime = getConnection().getTime();
 
  if (!pruefeGueltigkeit(currentTime))
  {
    throw new IllegalStateException(this + " ist nicht prfbar " +  
      "da (momentan) nicht gltig");
  }
  
  if (!pruefeLage())
  {
    throw new IllegalStateException(this + " ist nicht prfbar " +  
    "da Lage unspezifiziert");
  }
  
  Collection<SystemObject> result = null;
  SystemObject letztesSegment = getLetztesStrassenSegment();
  float laengeLetztesSegment = getLaengeStrassenSegment(letztesSegment) -
    getLage().offsetNach;
  
  _debug.config("Prfe Staus der Engstelle " + this + 
      " - im letzten Strassensegment " + letztesSegment.getName() + 
      " werden " + laengeLetztesSegment + "m benutzt");
  
  Iterator<Situation> it = _verwaltung.getVerwaltungsStaus().getVerwalteteObjekte().iterator();
  while (it.hasNext())
  {
    try
    {
      Situation stau = it.next();
      synchronized(stau)
      {
        if (!stau.pruefeGueltigkeit(currentTime))
        {
          continue;
        }
         
        if (!stau.pruefeLage())
        {
          continue;
        }
        
        Array segmente = stau.getLage().strassenSegmente;
        boolean stauGefunden = false;
        float stauLaengeNachEngstelle = 0;
        
        _debug.finer("Prfe Stau " + stau + 
            " mit " + segmente.getLength() + " Strassensegment(en)");
        
        for (int i = 0; i < segmente.getLength(); i+=1)
        {
          Data item = segmente.getItem(i);
          SystemObject segment = item.asReferenceValue().getSystemObject();
          _debug.finest("Prfe Stau-Strassensegment " + segment.getName()); 
        
          if (letztesSegment.getId() == segment.getId()) 
          {   
            // Falls Endsegment auch im Stau vorkommt
            // Setze die Variable stauLaengeNachEngstelle auf den Anfangspunkt des Staus
            // Korrigiere ggf. Offset
            stauLaengeNachEngstelle = laengeLetztesSegment * (-1);
            if (i == 0)
            {
              stauLaengeNachEngstelle += stau.getLage().offsetVor; 
            }
            if (stauLaengeNachEngstelle < 0)
            {
              stauGefunden = true;
              _debug.config("Stau " + stau + " erstreckt sich auf die letzten " +
                  (stauLaengeNachEngstelle * (-1)) + "m des Endsegments " + 
                  letztesSegment.getName() + " von Engstelle " + this);
            }
          }
          if (stauGefunden)
          {
            stauLaengeNachEngstelle += getLaengeStrassenSegment(segment);
            if (i == segmente.getLength() - 1)
            {
              stauLaengeNachEngstelle -= stau.getLage().offsetNach;
            }
            if (stauLaengeNachEngstelle > maxLaengeStauUeberBaustellenEnde)
            {
              break;
            }
          }
                    
        }
        if (stauGefunden && stauLaengeNachEngstelle > maxLaengeStauUeberBaustellenEnde)
        {
          if (result == null)
          {
            result = new ArrayList<SystemObject>();
          }
          result.add(stau.getObjekt());
          _debug.info("Stau " + stau + " geht mehr als " + 
              maxLaengeStauUeberBaustellenEnde + "m (" + 
              stauLaengeNachEngstelle + "m) ber das Ende von " +
              this + " hinaus");
        }
        else
        {
          setStauMeldungInaktiv(stau.getObjekt());
        }
      }
    }
    catch(ConcurrentModificationException e)
    {
      it = _verwaltung.getVerwaltungsStaus().getVerwalteteObjekte().iterator();
      _debug.fine("nderung der Staumenge whrend Prfung");
    }
  }  
  
  if (result == null)
  {  
    result = Collections.emptyList();
  }  
  
  return result;
}
  
  
  
  /**
   * Berechnet die Lnge eines Strassensegments
   * 
   * @param ss das Strassensegment, dessen Lnge berechnet werden soll
   * @return die Lnge des Strassensegments
   * @throws IllegalArgumentException falls ss nicht vom Typ "typ.straenSegment" ist
   */
  private float getLaengeStrassenSegment(SystemObject ss)
  {
    if (!ss.isOfType("typ.straenSegment"))
    {
      throw new IllegalArgumentException("Objekt vom Typ typ.straenSegment erwartet");
    }
    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.straenSegment");
    return ss.getConfigurationData(atg).getItem("Lnge").asScaledValue().floatValue();
  }
  
  /**
   * Fhrt die Engpaprfung fr diese Engstelle durch.
   * Die Verkehrsstrke am nchsten MQ stromabwrts hinter der Engstelle (max bis zum/zur nchsten
   * Kreuz/Dreieck/Anschlusstelle) wird mit dem  Produkt aus Restkapazitt und Parameter 
   * <code>FaktorVerkehrsStaerkeStromAbwaerts</code> verglichen. Die MQ unmittelbar stromabwrts 
   * hinter der Baustelle, bei denen das Produkt aus Parameterwert und Restkapazitt von der Verkehrsstrke
   * berschritten wird, werden zurckgemeldet.
   * 
   * @return Eine Sammlung von <code>ErgebnisEngpassPruefung</code> der MQ unmittelbar stromabwrts hinter der 
   * Baustelle, bei denen das Produkt aus Parameterwert und Restkapazitt von der Verkehrsstrke 
   * berschritten wird. Eine leere Sammlung bedeutet dabei ein positives, eine nichtleere Sammlung ein 
   * negatives Ergebnis der Plausibilittsprfung. 
   * @throws IllegalStateException wenn die Prfung nicht durchgefhrt werden kann, aufgrund
   * etwa nichtversorgter Parameter, Baustelle nicht gltig etc.
   */
  public synchronized Collection<ErgebnisEngpassPruefung> pruefeEngpass()
    throws IllegalStateException
  {
    float faktorVerkehrsStaerkeStromAbwaerts = getFaktorVerkehrsStaerkeStromAbwaerts();
    if (faktorVerkehrsStaerkeStromAbwaerts < 0)
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
        "da Parameter \"FaktorVerkehrsStaerkeStromAbwaerts\" nicht versorgt");
    }
        
    long currentTime = getConnection().getTime();

    if (!pruefeGueltigkeit(currentTime))
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
        "da (momentan) nicht gltig");
    }
 
    if (!pruefeLage())
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +  
        "da Lage unspezifiziert");
    }
    
    long engPassKapazitaet = getEngpassKapazitaet();
    if (engPassKapazitaet < 0)
    {
      throw new IllegalStateException(this + " ist nicht prfbar " +
        "da Attribut \"EngpassKapazitaet\" nicht versorgt");
    }
    
    Collection<ErgebnisEngpassPruefung> result = null;
    
    Collection<SystemObject> messQuerschnitte = _verkehrsStaerkeStromAbwaerts.keySet();
    _debug.config("Prfe " + messQuerschnitte.size() + " MQ hinter der Engstelle " + this);
    
    for (SystemObject mq: messQuerschnitte)
    {
      _debug.fine("Prfe MessQuerschnitt " + mq.getPid() + " hinter der Engstelle " + this);
    }
    
    Iterator<SystemObject> it = messQuerschnitte.iterator();
    
    while(it.hasNext())
    {
      try
      {
        SystemObject mq = it.next();
        Verkehrsstaerke verkehrsStaerke = _verkehrsStaerkeStromAbwaerts.get(mq);
        if (verkehrsStaerke.isQkfzPlausibel())
        {  
          long vs = verkehrsStaerke.getVerkehrsstaerkeQKfz(); 
          if (vs > engPassKapazitaet * faktorVerkehrsStaerkeStromAbwaerts)
          {
            if (result == null)
            {
              result = new ArrayList<ErgebnisEngpassPruefung>();
            }
            result.add(new ErgebnisEngpassPruefung(mq, vs, engPassKapazitaet));
            _debug.info("Am MQ " + mq.getName() + " ist die Verkehrsstrke mit " + vs +
                "Fzg/h um mehr als Faktor " + faktorVerkehrsStaerkeStromAbwaerts +
                " grer als die Kapazitt " + engPassKapazitaet + "Fzg/h von " + this);
          }
          else
          {
            kapazitaetsMeldungen.setMeldungInaktiv(getIdKapazitaetsMeldung(mq));
          }
        }
        else
        {
          _debug.finer("Am MQ " + mq.getName() + " ist der Messwert nicht plausibel");
          continue;
        }
      }
      catch (ConcurrentModificationException e)
      {
        _debug.config("Prfe " + messQuerschnitte.size() + " MQ hinter der Engstelle " + this);
        it = messQuerschnitte.iterator();
      }
    }
    
    if (result == null)
    {  
      result = Collections.emptyList();
    }  
    
    return result;
  }
  
  /**
   * Enthlt fr jeden MQ stromabwrts unmittelbar hinter der Engstelle eine Instanz vom Typ 
   * <code>Verkehrsstaerke</code>,
   * Schlssel ist das MQ-Systemobjekt.
   */
  private Map<SystemObject, Verkehrsstaerke> _verkehrsStaerkeStromAbwaerts =
    new HashMap<SystemObject, Verkehrsstaerke>();
  
  @Override
  public void abmelden()
  {
    super.abmelden();
    bereinigeVerkehrsStaerke();
  }
  
  @Override
  public void update(ResultData[] resultData)
  {
    SystemObject altesLetzteStrassenSegment = getLetztesStrassenSegment();
    float alterOffset = getLage() != null ? getLage().offsetNach : -1;
    super.update(resultData);
    SystemObject letztesStrassenSegment = getLetztesStrassenSegment();
    float offset = getLage() != null ? getLage().offsetNach : -1;
    
    if (letztesStrassenSegment != altesLetzteStrassenSegment ||
        offset != alterOffset)
    {
      synchronized(_verkehrsStaerkeStromAbwaerts)
      {
        bereinigeVerkehrsStaerke();
        if (letztesStrassenSegment != null)
        {
          try
          {
//            long mqOffset = (long)(getLaengeStrassenSegment(letztesStrassenSegment) - offset);
            long mqOffset = (long)offset;
            String text = "";
            for (SystemObject mq: getMqStromAbwaerts(letztesStrassenSegment.getPid(), mqOffset))
            {
              _verkehrsStaerkeStromAbwaerts.put(mq, new Verkehrsstaerke(getConnection(), mq.getPid()));
              text += ("\n\t" + mq.getNameOrPidOrId());
            }
            _debug.config("Verwendete(r) MQ stromabwrts hinter " + this +":" + 
                (text.length() == 0 ? "\n\t<kein>" : text));
          }
          catch(RuntimeException e)
          {
            _debug.error("Fehler beim bestimmen der Messquerschnitte stromabwrts hinter "  + this, e);
          }
        }
      }  
    }
  }
  
  /**
   * Bestimmt die MQ stromabwrts unmittelbar hinter dem bergebenen Straensegment, 
   * unter bercksichtigung des Offsets
   * 
   * @param ssPid die Pid des Strassensegments, hinter dem die MQ bestimmt werden sollen
   * @param offset Offset am Ende des bergebenen Straensegments
   * 
   * @return Sammlung von MQ stromabwrts unmittelbar hinter dem bergebenen Straensegment 
   */
  protected Collection<SystemObject> getMqStromAbwaerts(String ssPid, long offset)
  {
    Konfiguration konf = Konfiguration.getInstanz();
    StrassenSegment ss = konf.getStrassenSegment(ssPid);
    
    if (ss == null)
    {
      throw new IllegalArgumentException("Strassensegment mit PID " + ssPid +
          /* " von " + this + */ " ist KEIN Netzbestandteil");
    }
    
    // Prfe aktuelles Straensegment
    if (ss.isInneresStrassenSegment())
    {
      InneresStrassenSegment iss = (InneresStrassenSegment)ss;
      StrassenKnoten kn = null;
      
      if(iss.getNachStrassenSegment() != null)
      {  
        kn = iss.getNachStrassenSegment().getVonKnoten();
      }
      else if (iss.getVonStrassenSegmemt() != null)
      {
        kn = iss.getVonStrassenSegmemt().getNachKnoten();
      }
      
      if (kn != null && (kn.isAutobahnAnschlussStelle() ||
                         kn.isAutobahnKreuz()||
                         kn.isAutobahnDreieck()))
      {
        // Prfung soll in Kreuz/Dreieck/Anschlusstelle erfolgen => Abbruch
        _debug.fine("Strassensegment mit PID " + ssPid + "liegt auf " +
          kn.getTyp() + " => keine MQ");
        return Collections.emptySet();
      }
    }
    
    // Prfe MQ auf aktuellem Straensegment
    MessQuerschnitt mq = ss.getMessQuerschnittNachOffset(offset);
    if (mq != null)
    {
      Collection<SystemObject> result = new HashSet<SystemObject>();
      result.add(mq.getSystemObjekt());
      return result;
    }
    else
    {
      // Prfe nachfolgende Strassensegmente
      Collection<StrassenSegment> nachfolger = ss.getNachfolgerAufStrasse();
      Collection<SystemObject> result = null;
      if (nachfolger != null)
      {
        for (StrassenSegment nf: nachfolger)
        {
          Collection<SystemObject> col = getMqStromAbwaerts(nf.getSystemObject().getPid(), 0);
          if (!col.isEmpty())
          {
            if (result == null)
            {
              result = new HashSet<SystemObject>();
              result.addAll(col);
            }
          }
        }
      }
      if (result != null)
      {  
        return result;
      }
      else
      {
        return Collections.emptySet();
      }
    }
  }
  
  /**
   * Bestimmt das letzte Strassensegment dieser Engstelle
   * 
   * @return das Systemobjekt des letzten Straensegments dieser Baustelle, 
   * oder null, wenn z. B. keine Strassensegmente parametriert
   */
  public SystemObject getLetztesStrassenSegment()
  {
    synchronized(this)
    {
      SystemObject letztesSegment = null;
      Lage lage = getLage();
      if (lage != null)
      {
        int anzSeg = lage.strassenSegmente.getLength();
        Data item = lage.strassenSegmente.getItem(anzSeg - 1);
        letztesSegment = item.asReferenceValue().getSystemObject();
      }
      return letztesSegment;
    }
  }
  
  /**
   * Instanz von <code>MeldungsBehandlung</code> zur Behandlung von Betriebsmeldungen zur 
   * Stauprfung dieser Engstelle
   */
  public final MeldungsBehandlung stauMeldungen = new MeldungsBehandlung();
  
  /**
   * Bildet eine Id zur Verwaltung von Betriebsmeldungen bezgl. der Stauprfung 
   * fr diese Engsstelle
   * 
   * @param stau der referenzierte Stau
   * @return die gebildete Id
   */
  public String getIdStauMeldung(SystemObject stau)
  {
    return getObjekt().getPid() + "." + stau.getPid();
  }
  
  /**
   * Instanz von <code>MeldungsBehandlung</code> zur Behandlung von Betriebsmeldungen zur 
   * Engpaprfung dieser Engstelle
   */
  public final MeldungsBehandlung kapazitaetsMeldungen = new MeldungsBehandlung();
  
  /**
   * Bildet eine Id zur Verwaltung von Betriebsmeldungen bezgl. der Engpaprfung 
   * fr diese Engsstelle
   * 
   * @param mq der referenzierte MQ
   * @return die gebildete Id
   */
  public String getIdKapazitaetsMeldung(SystemObject mq)
  {
    return getObjekt().getPid() + "." + mq.getPid();
  }
  
  /**
   * Setzt eine Betriebsmeldung bezgl. Stauprfung auf "inaktiv"
   * Trgt der Eigenschaft "Dynamisches Objekt" von Staus Rechnung
   * 
   * @param stau der referenzierte Stau
   * @return true, wenn die Meldung vorher aktiv war, false sonst
   */
  public boolean setStauMeldungInaktiv(SystemObject stau)
  {
    boolean result = stauMeldungen.setMeldungInaktiv(getIdStauMeldung(stau));
    if (result)
    {
      _verwaltung.getVerwaltungsStaus().abmeldenBereinigung(this, stau);
    }
    return result;
  }
  
  /**
   * Setzt eine Betriebsmeldung bezgl. Stauprfung auf "aktiv"
   * Trgt der Eigenschaft "Dynamisches Objekt" von Staus Rechnung
   * 
   * @param stau der referenzierte Stau
   * @return true, wenn die Meldung vorher inaktiv war, false sonst
   */
  public boolean setStauMeldungAktiv(SystemObject stau)
  {
    boolean result = stauMeldungen.setMeldungAktiv(getIdStauMeldung(stau));
    if (result)
    {
      _verwaltung.getVerwaltungsStaus().anmeldenBereinigung(this, stau);
    }
    return result;
  }
  
  /**
   * Gibt mit dem dynamischen Objekt vom Typ Stau verbundene Ressourcen frei.
   * 
   * @param stau der referenzierte Stau
   */
  public void bereinige(SystemObject stau)
  {
    setStauMeldungInaktiv(stau);
  }
  
  /**
   * Leert die Map <code>_verkehrsStaerkeStromAbwaerts</code> und gibt mit ihren Eintrgen
   * evtl. verbundene Ressourcen frei.
   */
  protected void bereinigeVerkehrsStaerke()
  {
    synchronized(_verkehrsStaerkeStromAbwaerts) 
    {
      for (SystemObject mq: _verkehrsStaerkeStromAbwaerts.keySet())
      {
        Verkehrsstaerke vs = _verkehrsStaerkeStromAbwaerts.get(mq);
        vs.dispose();
        kapazitaetsMeldungen.setMeldungInaktiv(getIdKapazitaetsMeldung(mq));
      }
      _verkehrsStaerkeStromAbwaerts.clear();
    }
  }
  
  /**
   * Klasse zur Meldungsbehandlung.
   * Eine Betriebsmeldung wird dann versendet, wenn eine Plausibilittsprfung 
   * ein negatives Ergebenis hat und eine entsprechende Meldung (mit dergleichen Id)
   * nicht aktiv ist.
   * Sobald eine Betriebsmeldung versendet wird, wird die Meldung mit dieser Id auf "aktiv" gesetzt.
   * Eine nachfolgende Prfung mit positivem Ergebnis setzt die Meldung wieder auf "inaktiv"
   *  
   * @author Stefan Sans
   *
   */
  public class MeldungsBehandlung
  {
    /**
     * Menge der aktiven Meldungen
     */
    private Set<String> _aktiveMeldungen = new HashSet<String>();
    
    /**
     * Setzt eine Meldung auf "inaktiv"
     * 
     * @param meldungsId die Id der zu bearbeitenden Meldung
     * @return true, wenn die Meldung vorher "aktiv" war, false sonst
     */
    public boolean setMeldungInaktiv(String meldungsId)
    {
      return _aktiveMeldungen.remove(meldungsId);
    }
    
    /**
     * Setzt eine Meldung auf "aktiv"
     * 
     * @param meldungsId die Id der zu bearbeitenden Meldung
     * @return true, wenn die Meldung vorher "inaktiv" war, false sonst
     */
    public boolean isMeldungAktiv(String meldungsId)
    {
      return _aktiveMeldungen.contains(meldungsId);
    }
    
    /**
     * Bestimmt, ob eine Meldung auf "aktiv" ist.
     * 
     * @param meldungsId die Id der zu bearbeitenden Meldung
     * @return true, wenn die Meldung "aktiv" ist, false sonst
     */
    public boolean setMeldungAktiv(String meldungsId)
    {
      return _aktiveMeldungen.add(meldungsId);
    }
  }
}
