/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.engvew.engvew;

import java.lang.Thread.UncaughtExceptionHandler;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.dambach.konfiguration.netz.Konfiguration;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.vew.engvew.pruefung.PruefEngstellen;
import de.bsvrz.vew.engvew.verwaltung.VerwaltungEngstellen;

/**
 * Applikationsklasse der SWE EngstellenVerwaltung.<br>
 * Fr das bergebeneVerkehrsModellNetz (und optional Konfigurationsobjekt 
 * vom Typ typ.engstellenVerwaltung und Zykluszeit ) wird eine Instanz von 
 * <code>VerwaltungEngstellen</code> erzeugt und initialisiert.
 * 
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.2 $ / $Date: 2009/09/23 13:55:48 $ / ($Author: Sans $)
 */
public class EngstellenVerwaltungApp  implements StandardApplication 
{
  /**
   * Debug-Logger fr Logging-Ausgaben
   */
  private Debug _debug;

   /**
   * -netz von Kommandozeile
   */
  private String _netzPid = null;
  
  /**
   * -konfigObjekt von Kommandozeile - optiona
   */
  private String _konfigPid = null;
  
  /**
   * Die verwalteteten VerkehrsModellNetze (inkl. Unternetze);
   */
  private VerwaltungEngstellen _engstVerw = null;
  
  /**
   * -zyklusZeit von Kommandozeile - optional
   */
  private int _zyklusZeit = 30;
  
  /**
   * Datenverteiler-Verbindung
   */
  private ClientDavInterface _connection;
  
   
  /**
   * Default- Konstruktor.<br>
   * Wird typischerweise von <code>main</code> aufgerufen
   */
  public EngstellenVerwaltungApp()
  {
    try
    {
      Thread.sleep(100);
    }
    catch (InterruptedException e)
    {
    }
  }
  /**
   * Extrahiert - nach Prfung - das Konfigurationsobjekt vom Typ typ.netz 
   * aus dem Attribut {@link #_netzPid}
   * @return das gefundene Konfigurationsobjekt
   * @throws IllegalArgumentException, falls Konfigurationsobjekt nicht gefunden oder 
   * vom falschen Typ
   */
  private ConfigurationObject getNetz()
  {
    ConfigurationObject netz = null;
    
    try
    {
      netz = (ConfigurationObject)_connection.getDataModel().getObject(_netzPid);
    }
    catch(ClassCastException e)
    {
      throw new IllegalArgumentException("Die PID \"" + _netzPid + 
      "\" bezeichnet kein KonfigurationsObjekt");
    }
    
    if (netz == null)
    {
      throw new IllegalArgumentException("Zur PID \"" + _netzPid + 
      "\" wurde kein Konfigurationsobjekt gefunden ");
    }
  
    if (!netz.isOfType("typ.netz"))
    {
      throw new IllegalArgumentException("Das als Netz fr die " +
          "EngstellenVerwaltung vorgesehene Konfigurationsobjekt \"" + netz.getPid() +  
          "\" ist nicht vom Typ typ.netz"); 
    }
    
    return netz;
  }
  
  /**
   * Extrahiert - nach Prfung - das Konfigurationsobjekt fr die Engstellenverwaltung
   * aus dem Attribut {@link #_konfigPid}
   * @return das Konfigurationsobjekt zur Engstellenverwaltung
   * @throws IllegalArgumentException, falls Konfigurationsobjekt nicht gefunden 
   * oder vom falschen Typ
   */
  private ConfigurationObject getKonfigurationsObjekt()
  {
    ConfigurationObject co = null;
    if (_konfigPid != null)
    {
      try
      {
        co = (ConfigurationObject)_connection.getDataModel().getObject(_konfigPid);
      }
      catch(ClassCastException e)
      {
        throw new IllegalArgumentException("Die PID \"" + _konfigPid + 
        "\" bezeichnet kein KonfigurationsObjekt");
      }
      if (co == null)
      {
        throw new IllegalArgumentException("Zur PID \"" + _konfigPid + 
        "\" wurde kein Konfigurationsobjekt gefunden ");
      }
    }
    else
    {
      co = _connection.getLocalConfigurationAuthority();
    }
    if (!co.isOfType("typ.engstellenVerwaltung"))
    {
      throw new IllegalArgumentException("Das als Konfigurationsobjekt fr die " +
          "EngstellenVerwaltung vorgesehene Objekt \"" + co.getPid() +  
          "\" ist nicht vom Typ typ.engstellenVerwaltung"); 
    }
    return co;
  }
  

  /**
   * berschriebene Methode von <code>StandardApplication</code>, 
   * die die speziellen Startparameter auswertet.<br>
   * Die Argumwente, die ausgewertet werden, sind durch die Attribute {@link #_netzPid}, 
   * {@link #_konfigPid}(optional) und {@link #_zyklusZeit}(optional) spezifiziert
   * 
   * @param argumentList
   *  siehe sys.funclib.application.StandardApplication#parseArguments(sys.funclib.ArgumentList)
   */
  public void parseArguments(ArgumentList argumentList)
  {
    _debug = Debug.getLogger();
   
    _netzPid = argumentList.fetchArgument("-netz=").asNonEmptyString();
   
    if (argumentList.hasArgument("-konfigObjekt"))
    {  
      _konfigPid = argumentList.fetchArgument("-konfigObjekt=").asNonEmptyString();
    }
        
     if (argumentList.hasArgument("-zyklus"))
     {  
       _zyklusZeit = argumentList.fetchArgument("-zyklus=").intValueBetween(1, 86400);
     }    
      
    argumentList.fetchUnusedArguments();
  }
  
    
  /**
   * berschriebene Methode von StandardApplication, die die Initialisierung durchfhrt.<br>
   * Es wird eine Instanz von <code>VerwaltungEngstellen</code> erzeugt und initialisiert
   * 
   * @param connection
   * siehe sys.funclib.application.StandardApplication#initialize(de.bsvrz.dav.daf.main.ClientDavInterface)
   */
  public void initialize(ClientDavInterface connection)
  {
    _connection = connection;
    try
    {
      Thread.setDefaultUncaughtExceptionHandler(
          new UncaughtExceptionHandler()
          {
            public void uncaughtException(Thread t, Throwable e)
            {
              _debug.error("Uncaught Exception " + e + "in Thread " + t +":\n", e);
            }
          }
      );
      
      MessageSender.getInstance().setApplicationLabel(
          _connection.getLocalApplicationObject().getName());
      
      ConfigurationObject netz = getNetz();
      
      // Initialisiere Objekte des/der Netzes/Netze
      Konfiguration.getInstanz().bestimmeObjekte(connection, netz.getPid(), "");
      
      // Initialisiere Verwaltung(en) der Engstellen
      _engstVerw = new VerwaltungEngstellen(connection, netz, getKonfigurationsObjekt());
      
      _debug.info("Zykluszeit fr Plausibilittsprfungen der Engstellen ist " +
          _zyklusZeit + " sec.");
      Thread.sleep(2000);
      new PruefEngstellen(_zyklusZeit, _engstVerw);
      _debug.info("SWE EngstellenVerwaltung initialisiert");
    }
    catch (Exception e)
    {
      //e.printStackTrace();
      _debug.error("Abbruch der SWE EngstellenVerwaltung whrend der Initialisierung", e);
      System.exit(1);
    }
  }

  
  /**
   * Programmeinstieg Engstellenverwaltung.<br>
   * 
   * @param arguments Kommandozeilenargumente, folgende speziellen Argumente werden ausgewertet:<p>
   * -netz: VerkehrsModellNetz, dessen Engstellen verwaltet und berwacht werden.<p>
   * -konfigObjekt: Konfigurationsobjekt fr die Verwaltung von Engstellen - muss vom Typ 
   * typ.engstellenVerwaltung sein. Optional, wenn nicht angegeben, wird das Konfigurationsobjekt des
   * lokalen Konfigurationsverantwortlichen genommen.<p>
   *-zyklusZeit: Zykluszeit fr die Plausibilittsprfungen der Engstellen (in s). Optional, wenn nicht angegeben, 
   * werden die Prfungen alle 30s ausgefhrt. 
   */
  public static void main(String[] arguments)
  {
    StandardApplicationRunner.run(new EngstellenVerwaltungApp(), arguments);
  }

}
